#ifndef _Q931_H_
#define _Q931_H_
/*
 * Q.931
 *
 */


/*
 * Call states
 */
enum {
	Q931_STATE_NULL = 0,
	Q931_STATE_CALL_INITIATED = 1,
	Q931_STATE_OVERLAP_SENDING = 2,
	Q931_STATE_OUTGOING_CALL_PROCEEDING = 3,
	Q931_STATE_CALL_DELIVERED = 4,
//	Q931_STATE_ = 5,
	Q931_STATE_CALL_PRESENT = 6,
	Q931_STATE_CALL_RECEIVED = 7,
	Q931_STATE_CONNECT_REQUEST = 8,
	Q931_STATE_INCOMING_CALL_PROCEEDING = 9,
	Q931_STATE_ACTIVE = 10,
	Q931_STATE_DISCONNECT_REQUEST = 11,
	Q931_STATE_DISCONNECT_INDICATION = 12,

	Q931_STATE_SUSPEND_REQUEST = 15,

	Q931_STATE_RESUME_REQUEST = 17,

	Q931_STATE_RELEASE_REQUEST = 19,

	Q931_STATE_OVERLAP_RECEIVING = 25,
};

/*
 * Call direction
 */
enum {
	Q931_INCOMING = 0,
	Q931_OUTGOING
};

/*
 * Trunk state
 */
enum {
	Q931_TRUNK_DOWN = 0,
	Q931_TRUNK_UP
};

/*
 * Timer default values
 */
#define Q931_T301_DEFAULT	180000		/*  3 min */
#define Q931_T302_DEFAULT	15000		/* 15 sec */
#define Q931_T303_DEFAULT	4000		/*  4 sec */
#define Q931_T304_DEFAULT	20000		/* 20 sec */
#define Q931_T305_DEFAULT	30000		/* 30 sec */
#define Q931_T306_DEFAULT	30000		/* 30 sec */
#define Q931_T307_DEFAULT	180000		/*  3 min */
#define Q931_T308_DEFAULT	4000		/*  4 sec */
#define Q931_T309_DEFAULT	60000		/* 60 sec */
#define Q931_T310_DEFAULT	10000		/* 10 sec */
#define Q931_T312_DEFAULT	6000		/*  t303 + 2 sec */
#define Q931_T314_DEFAULT	2000		/*  2 sec */
#define Q931_T316_DEFAULT	120000		/*  2 min */
#define Q931_T317_DEFAULT	0		/* ?????? */
#define Q931_T320_DEFAULT	30000		/* 30 sec */
#define Q931_T321_DEFAULT	30000		/* 30 sec */
#define Q931_T322_DEFAULT	4000		/*  4 sec */

/**
 * Q.931 Call
 */
struct q931_call {
	/* id == crv */
	int id;

	/* state */
	int state;

	/* direction of call */
	int direction;

	/* Timers */
	unsigned long t301;
	unsigned long t302;
	unsigned long t303;
	unsigned long t304;
	unsigned long t305;
	unsigned long t306;
	unsigned long t307;
	unsigned long t308;
	unsigned long t309;
	unsigned long t310;
	unsigned long t312;
	unsigned long t314;
	unsigned long t316;
	unsigned long t317;
	unsigned long t320;
	unsigned long t321;
	unsigned long t322;

	/* Counters */
	unsigned int t301counter;
	unsigned int t302counter;
	unsigned int t303counter;
	unsigned int t304counter;
	unsigned int t305counter;
	unsigned int t306counter;
	unsigned int t307counter;
	unsigned int t308counter;
	unsigned int t309counter;
	unsigned int t310counter;
	unsigned int t312counter;
	unsigned int t314counter;
	unsigned int t316counter;
	unsigned int t317counter;
	unsigned int t320counter;
	unsigned int t321counter;
	unsigned int t322counter;
};

#define Q931_MAX_CALLS		256
#define Q931_MAX_MSGSIZE	3000

/**
 * Q.931 Trunk
 */
struct q931_trunk {
	/* trunk state */
	int state;

	/* crv handling */
	int next_crv;

	/* calls */
	struct q931_call	*calls[Q931_MAX_CALLS];

	/* default parameters */
	unsigned long t301_default;
	unsigned long t302_default;
	unsigned long t303_default;
	unsigned long t304_default;
	unsigned long t305_default;
	unsigned long t306_default;
	unsigned long t307_default;
	unsigned long t308_default;
	unsigned long t309_default;
	unsigned long t310_default;
	unsigned long t312_default;
	unsigned long t314_default;
	unsigned long t316_default;
	unsigned long t317_default;
	unsigned long t320_default;
	unsigned long t321_default;
	unsigned long t322_default;

	/* flags (TODO...) */
	int auto_connect_ack;

	/* callbacks */
	unsigned long (getTimeCB *)(void);

	void *tx34_priv;
	int (tx34_cb *)(struct q931_trunk *trunk, unsigned char *msg, int size, void *priv);

	void *tx32_priv;
	int (tx32_cb *)(struct q931_trunk *trunk, int type, unsigned char *msg, int size, void *priv);
};

enum {
	/* Call establishment */
	Q931_MSG_ALERTING		= 0x01,
	Q931_MSG_CALL_PROCEEDING	= 0x02,
	Q931_MSG_CONNECT		= 0x07,
	Q931_MSG_CONNECT_ACKNOWLEDGE	= 0x0f,
	Q931_MSG_PROGRESS		= 0x03,
	Q931_MSG_SETUP			= 0x05,
	Q931_MSG_SETUP_ACKNOWLEDGE	= 0x0d,

	/* Call information phase */
	Q931_MSG_RESUME			= 0x26,
	Q931_MSG_RESUME_ACKNOWLEDGE	= 0x2e,
	Q931_MSG_RESUME_REJECT		= 0x22,
	Q931_MSG_SUSPEND		= 0x25,
	Q931_MSG_SUSPEND_ACKNOWLEDGE	= 0x2d,
	Q931_MSG_SUSPEND_REJECT		= 0x21,
	Q931_MSG_USER_INFORMATION	= 0x20,

	/* Call clearing */
	Q931_MSG_DISCONNECT		= 0x45,
	Q931_MSG_RELEASE		= 0x4d,
	Q931_MSG_RELEASE_COMPLETE	= 0x5a,
	Q931_MSG_RESTART		= 0x46,
	Q931_MSG_RESTART_ACKNOWLEDGE	= 0x4e,

	/* Misc messages */
	Q931_MSG_SEGMENT		= 0x60,
	Q931_MSG_CONGESTION_CONTROL	= 0x79,
	Q931_MSG_INFORMATION		= 0x7b,
	Q931_MSG_NOTIFY			= 0x6e,
	Q931_MSG_STATUS			= 0x7d,
	Q931_MSG_STATUS_ENQUIRY		= 0x75
};


/**
 * Q.931 Message
 */
struct q931_message {
	int type;
	int size;

	int crv;
	int crvflag;

	/* reference */
	struct q931_call *call;

	unsigned char *raw;
};


int q931_get_crv(struct q931_trunk *trunk)
{
	return ++trunk->next_crv;
}

/*
 * TODO: not perfect...
 */
struct q931_call *q931_get_call(struct q931_trunk *trunk, const int crv)
{
	if(crv < 0 || crv >= Q931_MAX_CALLS)
		return NULL;

	return trunk->calls[crv];
}

int q931_tx32(struct q931_trunk *trunk, int type, unsigned char *msg, int size)
{
	if(!trunk->tx32_cb)
		return -1;

	return trunk->tx32_cb(trunk, type, msg, size, trunk->tx32_priv);
}

/*
 *
 */
int q931_tx34(struct q931_trunk *trunk, struct q931_call *call, unsigned char *msg, int size)
{
	if(!trunk->tx34_cb)
		return -1;

	return trunk->tx34_cb(trunk, call, msg, size, trunk->tx32_priv);
}

int q931_init(struct q931_trunk *trunk)
{
	memset(trunk->calls, 0, sizeof(trunk->calls));

	trunk->state = Q931_TRUNK_DOWN;
	trunk->next_crv = 0;

	/* default timer values */
	trunk->t301_default = Q931_T301_DEFAULT;
	trunk->t302_default = Q931_T302_DEFAULT;
	trunk->t303_default = Q931_T303_DEFAULT;
	trunk->t304_default = Q931_T304_DEFAULT;
	trunk->t305_default = Q931_T305_DEFAULT;
	trunk->t306_default = Q931_T306_DEFAULT;
	trunk->t307_default = Q931_T307_DEFAULT;
	trunk->t308_default = Q931_T308_DEFAULT;
	trunk->t309_default = Q931_T309_DEFAULT;
	trunk->t310_default = Q931_T310_DEFAULT;
	trunk->t312_default = Q931_T312_DEFAULT;
	trunk->t314_default = Q931_T314_DEFAULT;
	trunk->t316_default = Q931_T316_DEFAULT;
	trunk->t317_default = Q931_T317_DEFAULT;
	trunk->t320_default = Q931_T320_DEFAULT;
	trunk->t321_default = Q931_T321_DEFAULT;
	trunk->t322_default = Q931_T322_DEFAULT;

	trunk->auto_connect_ack = 0;

	return 0;
}

int q931_start(struct q931_trunk *trunk)
{
	memset(trunk->calls, 0, sizeof(trunk->calls));

	trunk->state = Q931_TRUNK_UP;

	return 0;
}

int q931_stop(struct q931_trunk *trunk)
{
	trunk->state = Q931_TRUNK_DOWN;

	return 0;
}

int q931_decode_crv(struct q931_trunk *trunk, struct q931_message *mes, unsigned char *buf, int size, int *offset)
{
	unsigned char len = 0;
	int pos = *offset;
	int crv = 0;

	if(size < *offset)
		return 0;

	/* CRV length */
	len = buf[pos++] & 0x0f;

	switch(len) {
	case 1:
		mes->crvflag = (buf[pos]   & 0x80) >> 7;
		mes->crv     =  buf[pos++] & 0x7f;
		break;

	case 2:
		mes->crvflag = (buf[pos]   & 0x80) >> 7;
		mes->crv     = (buf[pos++] & 0x7f) << 8;
		mes->crv    |=  buf[pos++];
		break;

	default:
		break;
	}

	*offset = pos;

	return crv;
}

/**
 * q931_rx23
 * \brief	Receive message from layer2
 */
int q931_rx23(struct q931_trunk *trunk, int type, unsigned char *buf, int size)
{
	struct q931_message mes;
	struct q931_call *call = NULL;
	int l2offset = 4;	/* 4 bytes offset */
	int offset = 0;
	int res = 0;


	switch(type) {
	default:
		memset(mes, 0, sizeof(struct q931_message));

		mes.raw  = &buf[l2offset];
		mes.size = size - l2offset;

		q931_decode_crv(trunk, &mes, &buf[l2offset], size - l2offset, &offset);
		mes.type = buf[offset++] & 0x7f;

		/* try to find associated call (if any) */
		call = q931_get_call(trunk, mes->crv);
		if(call) {
			mes->call = call;
		}

		res = q931_handle_message(trunk, &mes);
		break;
	}

	return res;
}

int q931_rx43(struct q931_trunk *trunk, q931_message *mes)
{
	unsigned char buf[Q931_MAX_MSGSIZE];

	mes->raw = buf;

	return 0;
}

int q931_handle_SETUP(struct q931_trunk *trunk, struct q931_mes *mes)
{
	struct q931_call *call;

	if(mes->call) {
		// bad mojo
	}

	call = malloc(sizeof(struct q931_call));
	if(!call) {
		// handle error
		return -1;
	}
	memset(call, 0, sizeof(struct q931_call));

	call->state = Q931_STATE_CALL_PRESENT;
	call->id    = mes->crv;

	trunk->calls[mes->crv] = call;

	/* decode IEs */

	/* send indication to L4 */
	res = q931_tx34(trunk, call, mes);

#if 0
	if(trunk->auto_setup_ack) {

	}
#endif
	return res;
}

#define Q931_LAYER2_OFFSET	4

/**
 * TODO: use cause ie struct instead of int
 */
int q931_disconnect(struct q931_trunk *trunk, struct q931_call *call, int cause)
{
	struct q931_message mes;
	unsigned char buf[Q931_MAX_MSGSIZE];
	int size = sizeof(buf) - Q931_LAYER2_OFFSET;

	switch(call->state) {
	case Q931_STATE_CALL_INITIATED:
		{
			/* send DISCONNECT message */
			mes.type = Q931_MSG_DISCONNECT;

			/* encode message */
			//res = q931_encode_message(trunk, mes, buf + Q931_LAYER2_OFFSET, &size);

			/* send to l2 */
			res = q931_tx32(trunk, buf, size);

			/* stop t303 */
			Q931_TIMER_STOP(trunk, call, t303);

			/* start t305 */
			Q931_TIMER_START(trunk, call, t305);

			/* change state */
			q931_change_state(trunk, call, Q931_STATE_DISCONNECT_REQUEST);
		}
		break;

	case Q931_STATE_OVERLAP_SENDING:
		{
			struct q931_message resp;
			int res;

			resp.type = Q931_MSG_RELEASE;

			/* TODO: add cause IE (cause code 6) */

			/* send release request */
			res = q931_send_message(trunk, call, &resp);

			/* stop t304 */
			Q931_TIMER_STOP(trunk, call, t304);

			/* start t308 */
			Q931_TIMER_START(trunk, call, t308);

			/* change state */
			q931_change_state(trunk, call, Q931_STATE_RELEASE_REQUEST);
		}
		break;

	case Q931_STATE_OUTGOING_CALL_PROCEEDING:
		{
			struct q931_message resp;
			int res;

			/* stop t310 */
			Q931_TIMER_STOP(trunk, call, t310);

			resp.type = Q931_MSG_RELEASE;

			/* TODO: add cause IE (cause code 6) */

			/* send release request */
			res = q931_send_message(trunk, call, &resp);

			/* start t308 */
			Q931_TIMER_START(trunk, call, t308);

			/* change state */
			q931_change_state(trunk, call, Q931_STATE_RELEASE_REQUEST);			
		}
		break;

	}
}

int q931_handle_RELEASE_COMPLETE(struct q931_trunk *trunk, struct q931_mes *mes)
{
	q931_call *call = NULL;

	if(!mes->call) {
		// bad mojo
	}
	call = mes->call;

	/* decode IEs */

	switch(call->state) {
	case Q931_STATE_CALL_INITIATED:
		{
			/* stop t303 */
			Q931_TIMER_STOP(trunk, call, t303);

			/* more info indication */
			res = q931_tx34(trunk, call, mes);

			/* release crv */
			mes->call = NULL;
			trunk->calls[mes->crv] = NULL;
		}
		break;

	default:
		/* invalid */
		break;
	}

	if(!mes->call) {
		q931_destroy_call(trunk, &call);
	}

	return res;
}

int q931_handle_PROGRESS(struct q931_trunk *trunk, struct q931_mes *mes)
{
	q931_call *call = NULL;

	if(!mes->call) {
		// bad mojo
	}
	call = mes->call;

	/* decode IEs */

	switch(call->state) {
	case Q931_STATE_OVERLAP_DIAL:
		/* stop t304 */
		Q931_TIMER_STOP(trunk, call, t304);

		/* progress indication */
		res = q931_tx34(trunk, call, mes);
		break;

	case Q931_STATE_OUTGOING_CALL_PROCEEDING:
		/* stop t310 */
		Q931_TIMER_STOP(trunk, call, t310);

		/* progress indication */
		res = q931_tx34(trunk, call, mes);
		break;

	default:
		/* invalid */
		break;
	}

	return res;
}

int q931_handle_ALERTING(struct q931_trunk *trunk, struct q931_mes *mes)
{
	q931_call *call = NULL;

	if(!mes->call) {
		// bad mojo
	}
	call = mes->call;

	/* decode IEs */

	switch(call->state) {
	case Q931_STATE_CALL_INITIATED:
		/* stop t303 */
		Q931_TIMER_STOP(trunk, call, t303);

		/* alerting indication */
		res = q931_tx34(trunk, call, mes);

		/* change state */
		q931_change_state(trunk, call, Q931_STATE_CALL_DELIVERED);
		break;

	case Q931_STATE_OVERLAP_SENDING:
		/* stop t304 */
		Q931_TIMER_STOP(trunk, call, t304);

		/* alerting indication */
		res = q931_tx34(trunk, call, mes);

		/* change state */
		q931_change_state(trunk, call, Q931_STATE_CALL_DELIVERED);
		break;

	case Q931_STATE_OUTGOING_CALL_PROCEEDING:
		/* stop t310 */
		Q931_TIMER_STOP(trunk, call, t310);

		/* start t301 */
		Q931_TIMER_START(trunk, call, t301);

		/* alerting indication */
		res = q931_tx34(trunk, call, mes);

		/* change state */
		q931_change_state(trunk, call, Q931_STATE_CALL_DELIVERED);
		break;

	default:
		/* invalid */
		break;
	}
	return res;
}

int q931_handle_PROCEEDING(struct q931_trunk *trunk, struct q931_mes *mes)
{
	q931_call *call = NULL;

	if(!mes->call) {
		// bad mojo
	}
	call = mes->call;

	/* decode IEs */

	switch(call->state) {
	case Q931_STATE_CALL_INITIATED:
		/* stop t303 */
		Q931_TIMER_STOP(trunk, call, t303);

		/* start t310 */
		Q931_TIMER_START(trunk, call, t310);

		/* proceeding indication */
		res = q931_tx34(trunk, call, mes);

		/* change state */
		q931_change_state(trunk, call, Q931_STATE_OUTGOING_CALL_PROCEEDING);
		break;

	case Q931_STATE_OVERLAP_SENDING:
		/* stop T304 */
		Q931_TIMER_STOP(trunk, call, t304);

		/* start T310 */
		Q931_TIMER_START(trunk, call, t310);

		/* proceeding indication */
		res = q931_tx34(trunk, call, mes);

		/* change state */
		q931_change_state(trunk, call, Q931_STATE_OUTGOING_CALL_PROCEEDING);
		break;

	default:
		/* invalid */
		break;
	}
	return res;
}


int q931_send_message(struct q931_trunk *trunk, struct q931_call *call, struct q931_mes *mes)
{
	unsigned char buf[Q931_MAX_MSGSIZE];
	int size, l2offset = Q931_LAYER2_OFFSET;

	if(!trunk || !call || !mes)
		return -1;
/*
	type = Q921_DL_DATA;

	if(trunk->mode == Q931_NT && trunk->type == Q931_BRI_PTMP && mes->type == Q931_MSG_SETUP) {
		l2offset -= 1;
		type = Q921_DL_UNIT_DATA;
	}
*/
	/* encode message */
	//res = q931_encode(trunk, call, mes, buf, &size);

	/* send message */
	res = q931_tx32(trunk, type, buf + l2offset, size);

	return res;
}

int q931_handle_CONNECT(struct q931_trunk *trunk, struct q931_mes *mes)
{
	q931_call *call = NULL;

	if(!mes->call) {
		// bad mojo
	}
	call = mes->call;

	/* decode IEs */

	switch(call->state) {
	case Q931_STATE_CALL_INITIATED:
		/* stop t303 */
		Q931_TIMER_STOP(trunk, call, t303);

		/* connect indication */
		res = q931_tx34(trunk, call, mes);

		/* ack? */
		if(trunk->auto_connect_ack) {
			q931_message resp;

			resp.type = Q931_MSG_CONNECT_ACKNOWLEDGE;

			/* send connect ack */
			q931_send_message(trunk, call, &resp);
		}

		/* change state */
		q931_change_state(trunk, call, Q931_STATE_ACTIVE);
		break;

	case Q931_STATE_OVERLAP_SENDING:
		/* stop t304 */
		Q931_TIMER_STOP(trunk, call, t304);

		/* connect indication */
		res = q931_tx34(trunk, call, mes);

		/* ack? */
		if(trunk->auto_connect_ack) {
			q931_message resp;

			resp.type = Q931_MSG_CONNECT_ACKNOWLEDGE;

			/* send connect ack */
			q931_send_message(trunk, call, &resp);
		}

		/* change state */
		q931_change_state(trunk, call, Q931_STATE_ACTIVE);
		break;

	case Q931_STATE_OUTGOING_CALL_PROCEEDING:
		/* stop t310 */
		Q931_TIMER_STOP(trunk, call, t310);

		/* connect indication */
		res = q931_tx34(trunk, call, mes);

		/* ack? */
		if(trunk->auto_connect_ack) {
			q931_message resp;

			resp.type = Q931_MSG_CONNECT_ACKNOWLEDGE;

			/* send connect ack */
			q931_send_message(trunk, call, &resp);
		}

		/* change state */
		q931_change_state(trunk, call, Q931_STATE_ACTIVE);
		break;

	case Q931_STATE_CALL_DELIVERED:
		/* connect indication */
		res = q931_tx34(trunk, call, mes);

		/* ack? */
		if(trunk->auto_connect_ack) {
			q931_message resp;

			resp.type = Q931_MSG_CONNECT_ACKNOWLEDGE;

			/* send connect ack */
			q931_send_message(trunk, call, &resp);
		}

		/* change state */
		q931_change_state(trunk, call, Q931_STATE_ACTIVE);
		break;

	default:
		/* invalid */
		break;
	}
	return res;
}


int q931_handle_SETUP_ACKNOWLEDGE(struct q931_trunk *trunk, struct q931_mes *mes)
{
	q931_call *call = NULL;

	if(!mes->call) {
		// bad mojo
	}
	call = mes->call;

	/* decode IEs */

	switch(call->state) {
	case Q931_STATE_CALL_INITIATED:
		/* stop t303 */
		Q931_TIMER_STOP(trunk, call, t303);

		/* more info indication */
		res = q931_tx34(trunk, call, mes);

		/* start t304 */
		Q931_TIMER_START(trunk, call, t304);

		/* change state */
		q931_change_state(trunk, call, Q931_STATE_OVERLAP_SENDING);
		break;

	default:
		/* invalid */
		break;
	}
}

int q931_handle_message(struct q931_trunk *trunk, struct q931_mes *mes)
{
	int res = -1;

	/*
	 * handle message based on state etc.
	 */
	switch(mes->type) {
	case Q931_MSG_SETUP:
		if(mes->call) {
			/* uh oh */
		} else {
			res = q931_handle_SETUP(trunk, mes);
		}
		break;

	case Q931_MSG_SETUP_ACKNOWLEDGE:
		res = q931_handle_SETUP_ACKNOWLEDGE(trunk, mes);
		break;
	}

	return res;
}



int q931_change_state(struct q931_trunk *trunk, struct q931_call *call, int state)
{
	int oldstate;

	if(!trunk || !call || state < 0)
		return -1;

	oldstate = call->state;

	/*
	 * actions based on target state
	 */
	switch(state) {

	default:
		break;
	}

	call->state = state;

	return 0;
}

int q931_destroy_call(struct q931_trunk *trunk, q931_call **call)
{
	free(*call);
	call = NULL;
}

void q931_t303_expired(struct q931_trunk *trunk, q931_call *call)
{
	if(!call)
		return;

	Q931_TIMER_STOP(trunk, call, t303);

	switch(call->state) {
	case Q931_STATE_CALL_INITIATED:
		/* first timeout? */
		if(!call->t303counter) {
			q931_message mes;
			unsigned char buf[Q931_MAX_MSGSIZE];
			int size = sizeof(buf);

			/* resend SETUP */
			mes.type = Q931_MSG_SETUP;
			//res = q931_encode_message(trunk, mes, buf, &size);

			/* start t303 */
			Q931_TIMER_START(trunk, call, t303);

			/* increment counter */
			call->t303counter++;
		} else {
			/* setup error */
			q931_tx34(trunk, call, NULL);

			/* release crv */
			trunk->calls[call->id] = NULL;
			q931_destroy_call(trunk, &call);
		}
		break;

	default:
		break;
	}

	return;
}

void q931_t304_expired(struct q931_trunk *trunk, q931_call *call)
{
	if(!trunk || !call)
		return;

	Q931_TIMER_STOP(trunk, call, t304);

	switch(call->state) {
	case Q931_STATE_OVERLAP_SENDING:
		{
			struct q931_messages resp;
			int res;

			resp.type = Q931_MSG_DISCONNECT;

			/* TODO: add cause IE */

			/* send disconnect */
			res = q931_send_message(trunk, call, &resp);

			/* start T305 */
			Q931_TIMER_START(trunk, call, t305);

			/* setup error */
			res = q931_tx34(trunk, call, NULL);

			/* change state */
			q931_change_state(trunk, call, Q931_STATE_DISCONNECT_REQUEST);
		}
		break;

	default:
		break;
	}

	return;
}

void q931_t305_expired(struct q931_trunk *trunk, q931_call *call)
{
	if(!call)
		return;

	Q931_TIMER_STOP(trunk, call, t305);

	switch(call->state) {
	default:
		break;
	}

	return;
}

void q931_t308_expired(struct q931_trunk *trunk, q931_call *call)
{
	if(!call)
		return;

	Q931_TIMER_STOP(trunk, call, t308);

	switch(call->state) {
	default:
		break;
	}

	return;
}

void q931_t310_expired(struct q931_trunk *trunk, q931_call *call)
{
	if(!call)
		return;

	Q931_TIMER_STOP(trunk, call, t310);

	switch(call->state) {
	case Q931_STATE_OUTGOING_CALL_PROCEEDING:
		{
			struct q931_message resp;
			int res;

			resp.type = Q931_MSG_DISCONNECT;

			/* TODO: add cause IE (?) */

			/* send */
			res = q931_send_message(trunk, call, &resp);

			/* setup error */
			res = q931_tx34(trunk, call, NULL);

			/* change state */
			q931_change_state(trunk, call, Q931_STATE_DISCONNECT_REQUEST);			
		}
		break;

	default:
		break;
	}

	return;
}


#define Q931_TIMER_EXPIRE(tr, x, tm, v) 			\
	if((tr)->calls[x]->tm && (tr)->calls[x]->tm <= v) {	\
		q931_##tm_expired(tr, (tr)->calls[x]);		\
	}

#define Q931_TIMER_STOP(tr, c, tm)	\
	if(c && c->tm) {		\
		c->tm = 0;		\
	}

#define Q931_TIMER_START(tr, c, tm)				\
	if(c && && (tr)->getTimeCB && !c->tm) {			\
		c->tm = (tr)->getTimeCB() + (tr)->##tm_default;	\
	}


void q931_timer_tick(struct q931_trunk *trunk)
{
	unsigned long now = 0;

	if(!trunk || !trunk->getTimeCB)
		return;

	if(trunk->state == Q931_TRUNK_DOWN)
		return;

	now = trunk->getTimeCB();

	for(x=0; x<Q931_MAX_CALLS; x++) {
		if(!trunk->calls[x])
			continue;

		/* check timers */
		Q931_TIMER_EXPIRE(trunk, x, t301, now);
		Q931_TIMER_EXPIRE(trunk, x, t302, now);
		Q931_TIMER_EXPIRE(trunk, x, t303, now);
		Q931_TIMER_EXPIRE(trunk, x, t304, now);
		Q931_TIMER_EXPIRE(trunk, x, t305, now);
		Q931_TIMER_EXPIRE(trunk, x, t306, now);
		Q931_TIMER_EXPIRE(trunk, x, t307, now);
		Q931_TIMER_EXPIRE(trunk, x, t308, now);
		Q931_TIMER_EXPIRE(trunk, x, t309, now);
		Q931_TIMER_EXPIRE(trunk, x, t310, now);
		Q931_TIMER_EXPIRE(trunk, x, t312, now);
		Q931_TIMER_EXPIRE(trunk, x, t314, now);
		Q931_TIMER_EXPIRE(trunk, x, t316, now);
		Q931_TIMER_EXPIRE(trunk, x, t317, now);
		Q931_TIMER_EXPIRE(trunk, x, t320, now);
		Q931_TIMER_EXPIRE(trunk, x, t321, now);
		Q931_TIMER_EXPIRE(trunk, x, t322, now);
	}
}

#endif
